# -*- coding: utf-8 -*-
'''
AiiDA-FLEUR
'''

__copyright__ = u"Copyright (c), 2015-2017, Forschungszentrum Juelich GmbH, Germany. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__contributors__ = "Jens Broeder"
__paper__ = ""
__paper_short__ = ""

