{
    "conv_mode": "density",
    "distance_charge": 0.1406279038,
    "distance_charge_all": [
        61.1110641131,
        43.7556515683,
        ...
    ],
    "distance_charge_units": "me/bohr^3",
    "errors": [],
    "force_diff_last": "can not be determined",
    "force_largest": 0.0,
    "info": [],
    "iterations_total": 23,
    "last_calc_uuid": "b20b5b94-5d80-41a8-82bf-b4d8eee9bddc",
    "loop_count": 1,
    "material": "FePt2",
    "total_energy": -38166.176928494,
    "total_energy_all": [
        -38166.542950054,
        -38166.345602746,
        ...
    ],
    "total_energy_units": "Htr",
    "total_wall_time": 245,
    "total_wall_time_units": "s",
    "warnings": [],
    "workflow_name": "FleurScfWorkChain",
    "workflow_version": "0.4.0"
}
